import os.path
import pysam
import asyncio
from pysam.utils import SamtoolsError

import threading
import logging
from varcall.modules.logging import setup_logging
from pathlib import Path

from varcall.modules.exec.common  import get_input

setup_logging()

def get_basename_and_ext(path) -> tuple:
    filename = os.path.splitext(os.path.basename(path))
    return filename


def run_samtools_view(self):
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "sam_view_input_input")
    output_path = get_input(self, "sam_view_output_input")
    view_region = get_input(self, "sam_view_region_input")

    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Samtools view")
        return

    # WARN:
    # outputs output_file in workingdir
    # change to add working dir or find a way to get input prefix before output_file name
    if not output_path:
        input_path_tuple = get_basename_and_ext(input_path)
        last = len(input_path_tuple)-1
        output_path = f"{input_path_tuple[0]}.view{input_path_tuple[last]}"

    self.notify(f"Samtools view {str(input_path)} to {str(input_path)} ...", title="Samtools view")
    logging.info(f"Samtools view {str(input_path)} to {str(input_path)} ...")
    self.query_one("#sam_view_horizontal").add_class("running")
    threading.Thread(
        target=_run_samtools_view,
        args=(
            self,
            input_path,
            output_path,
            view_region,
        ),
    ).start()



def _run_samtools_view(self, input_file, output_file, view_region):

    try:
        samtools_view_cmd = f"samtools view {input_file} -o {output_file} region={view_region}"
        self.notify(samtools_view_cmd, title="Samtools view")
        logging.info("Running command: " + samtools_view_cmd)

        with pysam.AlignmentFile(input_file, "rb") as infile, pysam.AlignmentFile(
            output_file, "wb", header=infile.header
        ) as outfile:
            for read in infile.fetch(region=view_region) if view_region else infile:
                outfile.write(read)

        logging.info(f"Samtools view completed for {str(input_file)}")
        self.notify(f"Samtools view completed for {str(input_file)}", title="Samtools view")

    except SamtoolsError as e:
        logging.error(f"An error occurred during Samtools view: {e}")
        self.notify(
            f"An error occurred during Samtools view: {e}",
            severity="error",
            timeout=10.0,
            title="Samtools view",
        )

    finally:
        self.query_one("#sam_view_horizontal").remove_class("running")


def run_samtools_sort(self):
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "sam_sort_input_input")
    output_path = get_input(self, "sam_sort_output_input")

    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Samtools sort")
        return

    # WARN:
    # outputs output_file in workingdir
    # change to add working dir or find a way to get input prefix before output_file name
    if not output_path:
        input_path_tuple = get_basename_and_ext(input_path)
        last = len(input_path_tuple)-1
        output_path = f"{input_path_tuple[0]}.sorted{input_path_tuple[last]}"

    self.notify(f"Samtools sort {str(input_path)} to {str(input_path)} ...", title="Samtools sort")
    logging.info(f"Samtools sort {str(input_path)} to {str(input_path)} ...")
    self.query_one("#sam_sort_horizontal").add_class("running")
    threading.Thread(
        target=_run_samtools_sort,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()

def _run_samtools_sort(self, input_file, output_file):

    try:
        samtools_sort_cmd = f"samtools sort {input_file} -o {output_file}"
        self.notify(samtools_sort_cmd, title="Samtools view")
        logging.info("Running command: " + samtools_sort_cmd)
        pysam.sort("-o", output_file, input_file)


        
        logging.info(f"Samtools sort completed for {str(input_file)}")
        self.notify(f"Samtools sort completed for {str(input_file)}", title="Samtools sort")


    except SamtoolsError as e:
        logging.error(f"An error occurred during Samtools sort: {e}")
        self.notify(
            f"An error occurred during Samtools sort: {e}",
            severity="error",
            timeout=10.0,
            title="Samtools sort",
        )

    finally:
        self.query_one("#sam_sort_horizontal").remove_class("running")


def run_samtools_index(self):
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "sam_index_input_input")

    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Samtools index")
        return

    self.notify(f"Samtools index {str(input_path)} ...", title="Samtools index")
    logging.info(f"Samtools index {str(input_path)} ...")
    self.query_one("#sam_index_horizontal").add_class("running")
    threading.Thread(
        target=_run_samtools_index,
        args=(
            self,
            input_path,
        ),
    ).start()

def _run_samtools_index(self, input_file):


    try:
        samtools_index_cmd = f"samtools index {input_file}"
        self.notify(samtools_index_cmd, title="Samtools index")
        logging.info("Running command: " + samtools_index_cmd)

        pysam.index(input_file)

        logging.info(f"Samtools index completed for {str(input_file)}")
        self.notify(f"Samtools index completed for {str(input_file)}", title="Samtools index")

    except SamtoolsError as e:
        logging.error(f"An error occurred during Samtools index: {e}")
        self.notify(
            f"An error occurred during Samtools index: {e}",
            severity="error",
            timeout=10.0,
            title="Samtools index",
        )

    finally:
        self.query_one("#sam_index_horizontal").remove_class("running")



# 4. samtools flagstats equivalent in pysam
def flagstats_bam(input_file, out_file):
    """
    The flagstats_bam function generates a summary of alignment statistics
    for the input BAM file, such as total number of reads, mapped reads,
    and properly paired reads.
    This is equivalent to `samtools flagstats` and is useful for assessing
    the quality of alignments.
    """
    stats = pysam.flagstats(input_file)
    out_file = Path(out_file)
    out_file.touch(exist_ok=True)
    with open(out_file, "w") as outfile:
        outfile.write(str(stats))


# 5. samtools stats equivalent in pysam
def stats_bam(input_file, output_file):
    """
    The stats_bam function calculates various statistics from the input BAM file
    and writes the results to the output file. The statistics include
    information on read lengths, GC content, and quality scores, among others.
    This is equivalent to `samtools stats`, providing detailed insights
    into the sequencing data.
    """
    with open(output_file, "w") as out:
        stats = str(pysam.stats(input_file))
        out.write(stats)




import unittest
from varcall.modules.exec.common import Command

class TestCommand(unittest.TestCase):
    def test_download_cmd(self):
        cmd = Command(name="download", workingDir = "trial", first_path="https://example.com/file.txt", second_path="file.txt")
        commands = cmd.get_command()
        expected = "curl -L https://example.com/file.txt -o file.txt"
        self.assertEqual(commands, expected)

    def test_fastqc_cmd(self):
        cmd = Command(name="fastqc", workingDir = "trial", first_path="sample.fastq", second_path="results/")
        commands = cmd.get_command()
        expected = "fastqc sample.fastq -o results/"
        self.assertEqual(commands, expected)

    def test_bwa_mem_cmd(self):
        cmd = Command(
            name="bwa_mem",
            workingDir = "trial",
            first_path="ref_genome.fa",
            second_path="8",
            third_input="reads_1.fastq",
            fourth_input="reads_2.fastq",
            fifth_input="aligned.sam"
        )
        commands = cmd.get_command()
        expected = (
            "bwa mem -t 8 ref_genome.fa reads_1.fastq reads_2.fastq -o aligned.sam"
        )
        self.assertEqual(commands, expected)

    def test_samtools_view_cmd(self):
        cmd = Command(
            name="samtools_view",
            workingDir = "trial",
            first_path="input.bam",
            second_path="output.sam",
            third_input="chr1:1000-2000"  # Region to view
        )
        commands = cmd.get_command()
        expected = "samtools view -b -h input.bam -o output.sam region=chr1:1000-2000"
        self.assertEqual(commands, expected)

    def test_bcf_filter_cmd(self):
        cmd = Command(
            name="bcf_filter",
            workingDir = "trial",
            first_path="variants.bcf",
            second_path="filtered.bcf",
            third_input="QUAL<20"  # Filter expression
        )
        commands = cmd.get_command()
        expected = "bcftools filter -s LOWQUAL -e 'QUAL<20' variants.bcf -o filtered.bcf"
        self.assertEqual(commands, expected)

    def test_bcf_stats_cmd(self):
        cmd = Command(name="bcf_stats", workingDir = "trial", first_path="variants.bcf", second_path="stats.txt")
        commands = cmd.get_command()
        expected = "bcftools stats variants.bcf > stats.txt"
        self.assertEqual(commands, expected)

    def test_error(self):
        cmd = Command(name="no_cmd", workingDir = "trial", first_path="no_input", second_path="no_output")
        commands = cmd.get_command()
        expected = "Error: 'no_cmd' command not found."
        self.assertEqual(commands, expected)

if __name__ == "__main__":
    unittest.main()



import unittest
from varcall.modules.exec.common import get_basename_and_ext, get_file_extension

class TestBasename(unittest.TestCase):

    def test_two_out(self):
        commands = get_basename_and_ext("aligned.sam")
        expected = ("aligned", ".sam")
        self.assertEqual(commands, expected)

    def test_two_three(self):
        commands = get_basename_and_ext("aligned.sorted.sam")
        expected = ("aligned.sorted", ".sam")
        self.assertEqual(commands, expected)

    def test_none(self):
        commands = get_basename_and_ext("")
        expected = ("", "")
        self.assertEqual(commands, expected)

class TestExtension(unittest.TestCase):

    def test_two_out(self):
        commands = get_file_extension("aligned.sam")
        expected = ".sam"
        self.assertEqual(commands, expected)

    def test_two_three(self):
        commands = get_file_extension("aligned.sorted.sam")
        expected = ".sam"
        self.assertEqual(commands, expected)

    def test_none(self):
        commands = get_file_extension("")
        expected = ("")
        self.assertEqual(commands, expected)

if __name__ == "__main__":
    unittest.main()



import os
import threading
import subprocess
import logging
from varcall.modules.logging import setup_logging
from varcall.modules.exec.samtools  import flagstats_bam
from varcall.modules.exec.common  import get_input, find_matching_files

setup_logging()


def run_download(self):
    """
    Initiates the download process for a given URL.

    This function retrieves the URL from the input field, validates it,
    and starts a new thread to handle the download process.

    Args:
        self: The instance of the class calling this function.
    """
    download_url = get_input(self, "Input_url")
    if not download_url:
        self.notify("Please provide a valid url", severity="warning", title="Download")
        return
    elif not self.full_output_path:
        self.notify("Please provide a valid url", severity="warning", title="Download")
        return
    else:
        self.notify(f"Downloading {str(download_url)}...", title="Download")
        logging.info(f"Downloading {str(download_url)}...")
        self.query_one("#Download_options").add_class("running")
        threading.Thread(
            target=_run_download,
            args=(
                self,
                download_url,
            ),
        ).start()


def _run_download(self, download_url: str) -> None:
    """
    Executes the download command using subprocess.

    This function runs the curl command to download the file from the given URL
    and saves it to the specified output path.

    Args:
        self: The instance of the class calling this function.
        download_url (str): The URL from which to download the file.
    """
    try:
        os.makedirs(name=f"{self.workingDir}/data/{self.outputdir_name}", exist_ok=True)
        download_cmd = f"curl -L {str(download_url)} -o {self.full_output_path}"
        self.notify(download_cmd, title="Download")
        logging.info("Running command: " + download_cmd)
        subprocess.run(
            download_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"Downloaded {str(download_url)}")
        self.notify(f"Downloaded {str(download_url)}", title="Download")
        self.query_one("#Download_options").remove_class("running")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred while Downloading: {e.stderr}")
        self.notify(
            f"An error occurred while Downloading: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="Download",
        )
    finally:
        self.query_one("#Download_options").remove_class("running")


def run_FastQC(self):
    """
    Initiates the FastQC process for quality control of sequencing data.

    This function retrieves the input file path from the input field, validates it,
    and starts a new thread to handle the FastQC process.

    Args:
        self: The instance of the class calling this function.
    """
    # if not self.workingDir:
    #     self.workingDir = __file__
    input_path = get_input(self, "FastQC_input_input")
    output_path = get_input(self, "FastQC_output_input")

    if not input_path:
        input_path = find_matching_files(f"{self.workingDir}/data/reads/*")
    if not output_path:
        output_path = f"{self.workingDir}/results/fastqc/"

    self.notify(f"FastQC {str(input_path)}...", title="FastQC")
    logging.info(f"FastQC {str(input_path)}...")
    self.query_one("#FastQC_Horizontal").add_class("running")
    threading.Thread(
        target=_run_FastQC,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()


def _run_FastQC(self, input_path: str, output_path: str) -> None:
    """
    Executes the FastQC command using subprocess.

    This function runs the FastQC command to perform quality control on the input file
    and saves the results to the specified output directory.

    Args:
        self: The instance of the class calling this function.
        input_path (str): The path to the input file for FastQC.
    """
    try:
        FastQC_cmd = f"fastqc {str(input_path)} -o {str(output_path)}"
        self.notify(FastQC_cmd, title="FastQC")
        logging.info("Running command: " + FastQC_cmd)
        subprocess.run(
            FastQC_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"FastQC completed for {str(input_path)}")
        self.notify(f"FastQC completed for {str(input_path)}", title="FastQC")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during FastQC: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="FastQC",
        )

    finally:
        self.query_one("#FastQC_Horizontal").remove_class("running")


def run_MultiQC(self):
    """
    Initiates the MultiQC process for aggregating results from multiple tools.

    This function retrieves the input file path from the input field, validates it,
    and starts a new thread to handle the MultiQC process.

    Args:
        self: The instance of the class calling this function.
    """
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "MultiQC_input_input")
    output_path = get_input(self, "MultiQC_output_input")
    if not input_path:
        input_path = find_matching_files(f"{self.workingDir}/results/fastqc/")
    if not output_path:
        output_path = f"{self.workingDir}/results/multiqc/"

    self.notify(f"MultiQC {str(input_path)}...", title="MultiQC")
    logging.info(f"MultiQC {str(input_path)}...")
    self.query_one("#MultiQC_Horizontal").add_class("running")
    threading.Thread(
        target=_run_MultiQC,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()


def _run_MultiQC(self, input_path: str, output_path: str) -> None:
    """
    Executes the MultiQC command using subprocess.

    This function runs the MultiQC command to aggregate results from multiple tools
    and saves the results to the specified output directory.

    Args:
        self: The instance of the class calling this function.
        input_path (str): The path to the input directory for MultiQC.
    """
    try:
        if not self.workingDir:
            self.workingDir = __file__
        MultiQC_cmd = f"multiqc {str(input_path)} -o {str(output_path)}"
        self.notify(MultiQC_cmd, title="MultiQC")
        logging.info("Running command: " + MultiQC_cmd)
        subprocess.run(
            MultiQC_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"MultiQC completed for {str(input_path)}")
        self.notify(f"MultiQC completed for {str(input_path)}", title="MultiQC")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during MultiQC: {e.stderr}")
        self.notify(
            f"An error occurred during MultiQC: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="MultiQC",
        )

    finally:
        self.query_one("#MultiQC_Horizontal").remove_class("running")


def run_bwa_index(self):
    """
    Initiates the BWA indexing process for the reference genome.

    This function retrieves the input file path from the input field, validates it,
    and starts a new thread to handle the BWA indexing process.

    Args:
        self: The instance of the class calling this function.
    """
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "bwa_ref_input")
    if not input_path:
        input_path = find_matching_files(f"{self.workingDir}/data/reference/*")
    self.notify(f"bwa indexing {str(input_path)}...", title="bwa index")
    logging.info(f"bwa indexing {str(input_path)}...")
    self.query_one("#bwa_Horizontal").add_class("running")
    threading.Thread(
        target=_run_bwa_index,
        args=(
            self,
            input_path,
        ),
    ).start()


def _run_bwa_index(self, input_path: str) -> None:
    """
    Executes the BWA index command using subprocess.

    This function runs the BWA index command to index the reference genome
    and saves the results to the specified output directory.

    Args:
        self: The instance of the class calling this function.
        input_path (str): The path to the reference genome file for BWA indexing.
    """
    try:
        if not self.workingDir:
            self.workingDir = __file__
        bwa_index_cmd = f"bwa index {str(input_path)}"
        self.notify(bwa_index_cmd, title="bwa Index")
        logging.info("Running command: " + bwa_index_cmd)
        subprocess.run(
            bwa_index_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bwa Index completed for {str(input_path)}")
        self.notify(f"bwa Index completed for {str(input_path)}", title="bwa Index")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bwa Index: {e.stderr}")
        self.notify(
            f"An error occurred during bwa Index: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bwa Index",
        )

    finally:
        self.query_one("#bwa_Horizontal").remove_class("running")


def run_bwa_mem(self):
    """
    Initiates the BWA MEM process for aligning reads to the reference genome.

    This function retrieves the input file paths and number of threads from the input fields,
    validates them, and starts a new thread to handle the BWA MEM process.

    Args:
        self: The instance of the class calling this function.
    """
    if not self.workingDir:
        self.workingDir = __file__

    ref_path = get_input(self, "bwa_ref_input")
    read_path_1 = get_input(self, "bwa_reads_input_1")
    read_path_2 = get_input(self, "bwa_reads_input_2")
    output_path = get_input(self, "bwa_output_input")
    no_of_threads = get_input(self, "bwa_threads_input")

    if not (read_path_1 and read_path_2):
        self.notify("Enter reads file path", title="bwa mem", severity="warning")
        return
    if not ref_path:
        self.notify("Enter reference genome file path", title="bwa mem", severity="warning")
        return
    if not no_of_threads:
        no_of_threads = 4
    if not output_path:
        output_path = f"{self.workingDir}/results/sam/aligned.sam"

    self.notify( f"aligning reads {str(read_path_1)} {str(read_path_2)} to {str(ref_path)} using bwa mem", title="bwa mem",)
    logging.info( f"aligning reads {str(read_path_1)} {str(read_path_2)} to {str(ref_path)} using bwa mem")

    self.query_one("#bwa_Horizontal").add_class("running")

    threading.Thread(
        target=_run_bwa_mem,
        args=(
            self,
            ref_path,
            read_path_1,
            read_path_2,
            no_of_threads,
            output_path,
        ),
    ).start()


def _run_bwa_mem(
    self, ref_path: str, read_path_1: str, read_path_2: str, no_of_threads: str, output_path) -> None:
    """
    Executes the BWA MEM command using subprocess.

    This function runs the BWA MEM command to align the reads to the reference genome
    and saves the results to the specified output directory.

    Args:
        self: The instance of the class calling this function.
        ref_path (str): The path to the reference genome file for BWA MEM.
        read_path (str): The path to the reads file for BWA MEM.
        no_of_threads (str): The number of threads to use for BWA MEM.
    """
    try:
        if not self.workingDir:
            self.workingDir = __file__
        bwa_mem_cmd = f"bwa mem -t {str(no_of_threads)} {str(ref_path)} {str(read_path_1)} {str(read_path_2)} -o {str(output_path)}"
        self.notify(bwa_mem_cmd, title="bwa mem")
        logging.info("Running command: " + bwa_mem_cmd)
        subprocess.run(
            bwa_mem_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(
            f"Completed aligning reads {str(read_path_1)} {str(read_path_2)} to {str(ref_path)} using bwa mem"
        )
        self.notify(
            f"Completed aligning reads {str(read_path_1)} {str(read_path_2)} to {str(ref_path)} using bwa mem",
            title="bwa mem",
        )

        flagstats_bam(
            f"{self.workingDir}/results/sam/aligned.sam",
            f"{self.workingDir}/results/sam/aligned.sam.flagstats",
        )

        MultiQC_cmd = f"multiqc {self.workingDir}/results/sam/aligned.sam.flagstats -o {self.workingDir}/results/multiqc/sam"
        subprocess.run(
            [MultiQC_cmd],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            shell=True,
        )
    except subprocess.CalledProcessError as e:
        logging.error(f"Error aligning using bwa mem: {e.stderr}")
        self.notify(
            f"Error aligning using bwa mem: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bwa mem",
        )

    finally:
        self.query_one("#bwa_Horizontal").remove_class("running")


# def run_Trimmomatic(self):
#     input_path = self.query_one("#bwa_ref_Input", Input)
#     input_path = input_path.value.strip()
#     if not input_path:
#         input_path = f"{self.workingDir}/data/reference/*"
#     self.notify(f"bwa indexing {str(input_path)}...", title="bwa index")
#     logging.info(f"bwa indexing {str(input_path)}...")
#     self.query_one("#bwa_Horizontal").add_class("running")
#     threading.Thread(
#         target=_run_Trimmomatic,
#         args=(
#             self,
#             input_path,
#         ),
#     ).start()
#
#
# def _run_Trimmomatic(self, input_path: str) -> None:
#     try:
#         Trimmomatic_cmd = (f"trimmomatic SE {input_fastq} {output_fastq} ILLUMINACLIP:TruSeq3-SE.fa:2:30:10 LEADING:3 TRAILING:3 SLIDINGWINDOW:4:15 MINLEN:36")
#         bwa_index_cmd = f"bwa index {str(input_path)}"
#         self.notify(bwa_index_cmd, title="bwa Index")
#         logging.info("Running command: " + bwa_index_cmd)
#         subprocess.run(
#             [Trimmomatic_cmd],
#             check=True,
#             stdout=subprocess.PIPE,
#             stderr=subprocess.PIPE,
#             text=True,
#             shell=True,
#         )
#         logging.info(f"bwa Index completed for {str(input_path)}")
#         self.notify(f"bwa Index completed for {str(input_path)}", title="bwa Index")
#         self.query_one("#bwa_Horizontal").remove_class("running")
#     except subprocess.CalledProcessError as e:
#         logging.error(f"An error occurred during bwa Index: {e.stderr}")
#         self.notify(
#             f"An error occurred during bwa Index: {e.stderr}",
#             severity="error", timeout=10.0,
#             title="bwa Index",
#         )



######################################################################
# Main app information.
__author__ = "Vidyasagar"
__copyright__ = "Copyright 2024, Vidyasagar"
__credits__ = ["Vidyasagar"]
__maintainer__ = "Vidyasagar"
__version__ = "0.0.2"
__licence__ = "MIT"

##############################################################################
# Local imports.
from .logging import setup_logging

from .exec.common import get_input, find_matching_files, file_suggester, get_basename_and_ext, get_file_extension
from .exec.hometab  import run_FastQC, run_MultiQC, run_bwa_mem, run_download, run_bwa_index
from .exec.bcftools import run_bcf_call, run_bcf_norm, run_bcf_stats, run_bcf_filter, run_bcf_mpileup
from .exec.samtools  import run_samtools_view, run_samtools_sort, run_samtools_index

from .widgets.Home import HomeWidgets
from .widgets.Samtools import SamtoolsWidgets
from .widgets.Bcftools import BcftoolsWidgets
from .widgets.Pipeline import PipelineWidgets

from .Help.Help import HelpMarkdown

##############################################################################
# Export the imports.
__all__ = [
    "setup_logging",

    "get_input", "find_matching_files", "file_suggester", "get_basename_and_ext", "get_file_extension",
    "run_bcf_call", "run_bcf_norm", "run_bcf_stats", "run_bcf_filter", "run_bcf_mpileup", 
    "run_bwa_index", "run_bwa_mem", "run_download", "run_MultiQC", "run_FastQC",
    "run_samtools_view", "run_samtools_sort", "run_samtools_index",

    "HomeWidgets",
    "SamtoolsWidgets",
    "BcftoolsWidgets",
    "PipelineWidgets",

    "HelpMarkdown",
]

### __init__.py ends here



import glob
from datetime import datetime
import os.path

# from ...src.main import Varcall
import logging
import subprocess
import threading
from pathlib import Path

from varcall.modules.logging import setup_logging
from textual.suggester import Suggester
from textual.widgets import Input

setup_logging()


def get_random_file_name_current(ext:str) -> str:

    now = datetime.now()
    date_time = now.strftime("%H%M_%d%m%Y")
    path = os.getcwd()
    path = f"{path}/{date_time}.{ext}"

    return path

def default_file_name(self, tool:str, input:str , has_cwd_set: bool) -> str:

    input_path_tuple = get_basename_and_ext(input)
    f"{input_path_tuple[0]}.sorted{input_path_tuple[1]}"

    if has_cwd_set:
        path = self.workingDir

    path = os.getcwd()
    now = datetime.now()
    date_time = now.strftime("%H%M_%d%m%Y")

    match tool:
        case "download":
            return f"{path}/{date_time}.download"

        case "fastqc":
            return path+"results/fastqc/"
        case "multiqc":
            return path+f"results/multiqc{date_time}/"
            
        case "bwa_index":
            return path+"data/reference/"
        case "bwa_mem":
            return path+f"results/sam/aligned{date_time}.sam"

        case "samtools_sort":
            return path+f"results/sam/aligned{date_time}.sorted.sam"
        case "samtools_view":
            return path+f"results/bam/aligned{date_time}.bam"
        case "samtools_index":
            return path+"results/sam/"

        case "bcf_call":
            return path+f"results/bcf/{date_time}.call.bcf"
        case "bcf_filter":
            return path+f"results/multiqc/"
        case "bcf_norm":
            return path+f"results/multiqc/"
        case "bcf_stats":
            return path+f"results/multiqc/"


    return path

def get_output_name(self, tool: str, input) -> str:
    if not self.workingDir:
        return default_file_name(self, tool, input, True)
    else:
        return default_file_name(self, tool, input, False)



class FileSuggester(Suggester):
    async def get_suggestion(self, value: str) -> str | None:
        """
        Provides file path suggestions based on the input value.
        """
        # Handle absolute paths
        if value.startswith("/"):
            return " "
        # Handle invalid pattern '**' anywhere in the input
        if "**" in value:
            return " "
        path = next(Path().glob(f"{value}*"), None)
        return str(path) if path else None


file_suggester = FileSuggester(use_cache=False, case_sensitive=True)


def get_basename_and_ext(path) -> tuple:
    """returns a tuple with basename and extension (has only two elements) e.g: ("basename", ".ext")
    returns ("", "") if input is false"""
    filename = os.path.splitext(os.path.basename(path))
    return filename

def get_file_extension(path) -> str:
    filename = os.path.splitext(os.path.basename(path))
    return filename[-1]


def find_matching_files(path_pattern: str) -> str:
    """Find all files matching the given wildcard path pattern."""
    matching_files = glob.glob(path_pattern)
    return " ".join(matching_files)


def get_input(self, input_widget_id) -> str:
    try:
        raw_input = str(self.query_one(f"#{input_widget_id}", Input).value).strip()
        return find_matching_files(raw_input)
    except Exception:
        return ""


class Command:
    def __init__(self, workingDir: str, name: str, first_path: str, second_path: str = "", third_input: str = "", fourth_input: str = "", fifth_input = None) -> None:
        self.name = name
        self.workingDir = workingDir
        self.first_path = first_path
        self.second_path = second_path
        self.third_input = third_input
        self.fourth_input = fourth_input
        self.fifth_input = fifth_input

    def get_command(self) -> str:
    # Generate commands dynamically based on instance variables
        commands = {
            
            "download": f"curl -L {self.first_path} -o {self.second_path or get_output_name(self, "download", self.first_path)}",

            "fastqc": f"fastqc {self.first_path} -o {self.second_path or get_output_name(self, "fastqc", self.first_path)}",
            "multiqc": f"multiqc {self.first_path} -o {self.second_path or get_output_name(self, "multiqc", self.first_path)}",

            "bwa_index": f"bwa index {self.first_path or get_output_name(self, "bwa_index", self.first_path)}",
            "bwa_mem": f"bwa mem -t {self.second_path or 4} {self.first_path} {self.third_input} {self.fourth_input} -o {self.fifth_input or get_output_name(self, "bwa_mem", self.first_path)}",

            "samtools_sort": f"samtools sort {self.first_path} -o {self.second_path or get_output_name(self, "samtools_sort", self.first_path)}",
            "samtools_view": f"samtools view -b -h {self.first_path} -o {self.second_path or get_output_name(self, "samtools_view", self.first_path)} region={self.third_input}",
            "samtools_index": f"samtools index {self.first_path or get_output_name(self, "samtools_index", self.first_path)}",

            "bcf_call": f"bcftools call -mv -Oz -o {self.second_path or get_output_name(self, "bcf_call", self.first_path)} {self.first_path}",
            "bcf_filter": f"bcftools filter -s LOWQUAL -e '{self.third_input}' {self.first_path} -o {self.second_path or get_output_name(self, "bcf_filter", self.first_path)}",
            "bcf_norm": f"bcftools norm -f {self.first_path} {self.second_path or get_output_name(self, "bcf_norm", self.first_path)}",
            "bcf_stats": f"bcftools stats {self.first_path} > {self.second_path or get_output_name(self, "bcf_stats", self.first_path)}"
        }
        return commands.get(self.name, f"Error: '{self.name}' command not found.")




def run_command(self, name: str, first_path : str, second_path : str = "", third_input: str = "", fourth_input: str = "", fifth_input: str = ""):

    _name = name.replace(" ", "_")

    first_path = get_input(self, f"{_name}_first_path")
    second_path = get_input(self, f"{_name}_second_path")
    third_input = get_input(self, f"{_name}_third_input")
    fourth_input  = get_input(self, f"{_name}_fourth_inpu")
    fifth_input = get_input(self, f"{_name}_fifth_input")

    if not first_path and (name != "fastqc" or "multiqc" or "bwa_index"):
        self.notify( "Please provide a valid path", severity="warning", title=name)
        return

    self.notify(f"{name} {str(first_path )}...", title=name)
    logging.info(f"{name} {str(first_path )}...")
    self.query_one(f"#{_name}_horizontal").add_class("running")
    cmd = Command(
        name,
        first_path,
        second_path,
        third_input,
        fourth_input,
        fifth_input,
    ).get_command()
    threading.Thread(
        target=_run_command,
        args=(
            self,
            cmd,
            name,
        ),
    ).start()


def _run_command(self, cmd: str, name:str) -> None:
    """ """
    _name = name.replace(" ", "_")
    try:
        self.notify(cmd, title=name)
        logging.info("Running command: " + cmd)
        subprocess.run(
            cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"{name} completed")
        self.notify(f"{name} completed", title=name)
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during {name}: {e.stderr}")
        self.notify(
            f"An error occurred during {name}: {e.stderr}",
            severity="error",
            timeout=10.0,
            title=name,
        )

    finally:
        self.query_one(f"#{_name}_horizontal").remove_class("running")



import unittest
from unittest.mock import patch
from varcall.modules.exec.common import find_matching_files

class TestFindMatchingFiles(unittest.TestCase):

    @patch("glob.glob")
    def test_find_matching_files_simple_pattern(self, mock_glob):
        # Mocking glob to return files matching a simple pattern
        mock_glob.return_value = ["file1.txt", "file2.txt"]
        
        result = find_matching_files("*.txt")
        self.assertEqual(result, "file1.txt file2.txt")

        mock_glob.assert_called_once_with("*.txt")

    @patch("glob.glob")
    def test_find_matching_files_no_match(self, mock_glob):
        # Mocking no matching files
        mock_glob.return_value = []

        result = find_matching_files("*.csv")
        self.assertEqual(result, "")

        mock_glob.assert_called_once_with("*.csv")

    @patch("glob.glob")
    def test_find_matching_files_nested_directories(self, mock_glob):
        # Mocking files in nested directories
        mock_glob.return_value = [
            "./trial/a/file1.txt", 
            "./trial/b/file2.txt"
        ]

        result = find_matching_files("./trial/*/file*.txt")
        self.assertEqual(result, "./trial/a/file1.txt ./trial/b/file2.txt")

        mock_glob.assert_called_once_with("./trial/*/file*.txt")

    @patch("glob.glob")
    def test_find_matching_files_partial_file_name(self, mock_glob):
        # Mocking files matching partial names
        mock_glob.return_value = [
            "./trial/reads/sample1.fastq", 
            "./trial/reads/sample1.log"
        ]

        result = find_matching_files("./trial/reads/sample1*")
        self.assertEqual(result, "./trial/reads/sample1.fastq ./trial/reads/sample1.log")

        mock_glob.assert_called_once_with("./trial/reads/sample1*")


    @patch("glob.glob")
    def test_find_matching_files_empty_string(self, mock_glob):
        # Mocking no pattern given (empty string)
        mock_glob.return_value = []

        result = find_matching_files("")
        self.assertEqual(result, "")

        mock_glob.assert_called_once_with("")

if __name__ == "__main__":
    unittest.main()



from textual.screen import ModalScreen
from textual.widgets import Button, Label
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual import on


class YesOrNo(ModalScreen):
    CSS = """
        YesOrNo{
            width: auto;
            height: auto;
            align: center middle;
            Container{
                width: auto;
                height: auto;
                padding: 1 2;
                margin: 2;
                background: $panel;
                }
            Label{
                width: 100%;
                height: auto;
                margin: 1;
                content-align: center middle;
                align: center middle;
                }
            #horizontal_yesorno{
                width: auto;
                height: auto;
                margin: 1;
                Button{
                    margin: 0 1 0 0;
                    }
                }
            }"""

    def __init__(self, Title: str = "Input") -> None:
        self.Title = Title
        super().__init__()

    def compose(self) -> ComposeResult:
        with Container(id="yesorno"):
            yield Label(self.Title)
            with Horizontal(id="horizontal_yesorno"):
                yield Button.success("Yes", id="yes")
                yield Button.error("No", id="no")

    @on(Button.Pressed)
    def close_modalscreen(self, event: Button.Pressed) -> None:
        self.dismiss(event.button.id == "yes")



import logging
import subprocess
from os import listdir
from webbrowser import open_new_tab

from textual import on
from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.reactive import reactive
from textual.widgets import (
    Button,
    Footer,
    Header,
    Input,
    Label,
    Select,
    TabbedContent,
    TabPane,
)

# Importing custom varcall.modules and functions
from varcall.modules.widgets.Bcftools import BcftoolsWidgets
from varcall.modules.exec.bcftools import run_bcf_call, run_bcf_filter, run_bcf_mpileup, run_bcf_stats, run_bcf_norm
from varcall.modules.exec.hometab import *  # noqa: F403
from varcall.modules.exec.samtools import *  # noqa: F403
from varcall.modules.Help.Help import HelpMarkdown
from varcall.modules.widgets.Home import HomeWidgets
from varcall.modules.logging import setup_logging
from varcall.modules.widgets.Pipeline import PipelineWidgets
from varcall.modules.widgets.Samtools import SamtoolsWidgets
from varcall.modules.widgets.YesOrNo import YesOrNo

# Setup logging configuration
setup_logging()


class Varcall(App[None]):
    # Key bindings for the application
    BINDINGS = [
        ("f1", "show_help", "Help"),
        ("q", "exit_app", "Exit App"),
        ("D", "toggle_dark_mode", "Dark/Light mode"),
        ("d", "run_download", "Download"),
        ("f", "run_fastqc", "Run FastQC"),
        ("m", "run_multiqc", "Run MultiQC"),
        ("i", "run_index", "Index reference genome"),
        ("a", "run_alignment", "Align reads"),
    ]
    CSS_PATH = "main.css"

    def on_mount(self) -> None:
        self.action_show_help()

    # Reactive variables to track state
    workingDir = reactive("")
    outputfile_name = reactive("")
    outputdir_name = reactive("")
    full_output_path = reactive("")

    def compose(self) -> ComposeResult:
        """
        Composes the UI layout of the application.

        This method sets up the main components of the UI, including the header,
        footer, and tabbed content for different functionalities.
        """
        with ScrollableContainer(id="ScrollableContainer"):
            yield Header()
            yield Footer(show_command_palette=True)
            with TabbedContent():
                with TabPane("Home", id="HomeTab"):
                    yield HomeWidgets()
                with TabPane("Samtools", id="SamtoolsTab"):
                    yield SamtoolsWidgets()
                with TabPane("Bcftools", id="BcftoolsTab"):
                    yield BcftoolsWidgets()
                with TabPane("Pipeline", id="Pipeline"):
                    yield PipelineWidgets()
                with TabPane("Help", id="HelpTab"):
                    yield HelpMarkdown()

    # Event handler for project name input submission
    @on(Input.Submitted, "#Input_Project_name")
    def update_working_dir(self, event: Input.Submitted) -> None:
        """
        Updates the working directory based on the project name input.

        This method is triggered when the project name input is submitted.
        It updates the working directory, creates necessary subdirectories,
        and notifies the user.

        Args:
            event (Input.Submitted): The event object containing the input value.
        """
        self.workingDir = str(event.input.value)
        self.update_full_output_path()
        self.notify(
            f"New Project Started and working directory updated to {self.workingDir}"
        )
        # Create necessary directories for the project
        makedir = f"mkdir -p {self.workingDir} {self.workingDir}/results {self.workingDir}/data {self.workingDir}/data/reads {self.workingDir}/data/reference {self.workingDir}/results/fastqc {self.workingDir}/results/multiqc {self.workingDir}/results/multiqc/sam {self.workingDir}/results/sam {self.workingDir}/results/bam {self.workingDir}/results/vcf {self.workingDir}/results/bcf"
        self.notify(makedir)
        try:
            subprocess.run(
                makedir.split(),
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
        except subprocess.CalledProcessError as e:
            logging.error(f"An error occurred: {e.stderr}")
            self.notify(f"An error occurred: {e.stderr}", severity="error")

    # Event handler for output file name input change
    @on(Input.Changed, "#Input_outputfile_name")
    def update_outputfile_name(self, event: Input.Changed) -> None:
        """
        Updates the output file name based on the input change.

        This method is triggered when the output file name input is changed.

        Args:
            event (Input.Changed): The event object containing the input value.
        """
        self.outputfile_name = event.input.value
        self.update_full_output_path()

    # Event handler for output directory selection change
    @on(Select.Changed, "#Select_outputdir")
    def update_outputdir_name(self, event: Select.Changed) -> None:
        """
        Updates the output directory name based on the selection change.

        This method is triggered when the output directory selection is changed.
        It updates the output directory name and the full output path.

        Args:
            event (Select.Changed): The event object containing the selected value.
        """
        self.outputdir_name = event.select.value
        self.update_full_output_path()

    # Event handler for download button press
    @on(Button.Pressed, "#Download_button")
    def call_run_download(self) -> None:
        """
        Initiates the download process when the download button is pressed.

        This method is triggered when the download button is pressed.
        It calls the run_download function to start the download process.
        """
        run_download(self)  # noqa: F405

    # Event handler for FastQC button press
    @on(Button.Pressed, "#FastQC_Button")
    def call_run_FastQC(self) -> None:
        """
        Initiates the FastQC process when the FastQC button is pressed.

        This method is triggered when the FastQC button is pressed.
        It calls the run_FastQC function to start the FastQC process.
        """
        run_FastQC(self)  # noqa: F405

    # Event handler for MultiQC button press
    @on(Button.Pressed, "#MultiQC_Button")
    def call_run_MultiQC(self) -> None:
        """
        Initiates the MultiQC process when the MultiQC button is pressed.

        This method is triggered when the MultiQC button is pressed.
        It calls the run_MultiQC function to start the MultiQC process.
        """
        run_MultiQC(self)  # noqa: F405

    # Event handler for BWA index button press
    @on(Button.Pressed, "#bwa_index_Button")
    def call_run_bwa_index(self) -> None:
        """
        Initiates the BWA indexing process when the BWA index button is pressed.

        This method is triggered when the BWA index button is pressed.
        It calls the run_bwa_index function to start the BWA indexing process.
        """
        run_bwa_index(self)  # noqa: F405

    # Event handler for BWA align button press
    @on(Button.Pressed, "#bwa_mem_Button")
    def call_run_bwa_mem(self) -> None:
        """
        Initiates the BWA MEM process when the BWA align button is pressed.

        This method is triggered when the BWA align button is pressed.
        It calls the run_bwa_mem function to start the BWA MEM process.
        """
        run_bwa_mem(self)  # noqa: F405

    # Event handler to view FastQC results
    @on(Button.Pressed, "#view_FastQC_res")
    def view_FastQC_res(self) -> None:
        """
        Opens the FastQC results in a web browser.

        This method is triggered when the view FastQC results button is pressed.
        It opens the FastQC results HTML files present
        in the workingDir/results/fastqc directory in a new browser tab.
        """
        if not self.workingDir:
            self.workingDir = __file__
        dir_path = f"{self.workingDir}/results/fastqc/"
        for file in listdir(dir_path):
            if file.endswith(".html"):
                open_new_tab(f"{dir_path}/{file}")

    # Event handler to view MultiQC results
    @on(Button.Pressed, "#view_MultiQC_res")
    def view_MultiQC_res(self) -> None:
        """
        Opens the MultiQC results in a web browser.

        This method is triggered when the view MultiQC results button is pressed.
        It opens the MultiQC results HTML files present
        in the workingDir/results/multiqc directory in a new browser tab.
        """
        if not self.workingDir:
            self.workingDir = __file__
        dir_path = f"{self.workingDir}/results/multiqc/"
        for file in listdir(dir_path):
            if file.endswith(".html"):
                open_new_tab(f"{dir_path}/{file}")

    @on(Button.Pressed, "#view_bwa_res")
    def view_bwa_res(self) -> None:
        """
        This method is triggered when the view bwa results button is pressed.
        Executes flagstats_bam() and multiqc on the result file and then,
        Opens the it in a web browser.
        """

        if not self.workingDir:
            self.workingDir = __file__
        dir_path = f"{self.workingDir}/results/multiqc/sam/"
        for file in listdir(dir_path):
            if file.endswith(".html"):
                open_new_tab(f"{dir_path}/{file}")

    @on(Button.Pressed, "#sam_view_button")
    def call_sam_view(self):
        run_samtools_view(self)

    @on(Button.Pressed, "#sam_sort_button")
    def call_sam_sort(self):
        run_samtools_sort(self)

    @on(Button.Pressed, "#sam_index_button")
    def call_sam_index(self):
        run_samtools_index(self)

    @on(Button.Pressed, "#bcf_mpileup_button")
    def call_bcf_mpileup(self):
        run_bcf_mpileup(self)

    @on(Button.Pressed, "#bcf_call_button")
    def call_bcf_call(self):
        run_bcf_call(self)

    @on(Button.Pressed, "#bcf_filter_button")
    def call_bcf_filter(self):
        run_bcf_filter(self)

    @on(Button.Pressed, "#bcf_stats_button")
    def call_bcf_stats(self):
        run_bcf_stats(self)

    @on(Button.Pressed, "#bcf_norm_button")
    def call_bcf_norm(self):
        run_bcf_norm(self)


    # Handels the key press for Toggle dark mode
    def action_toggle_dark_mode(self) -> None:
        self.dark = not self.dark

    # Handels the key press for Run download action
    def action_run_download(self) -> None:
        self.call_run_download()

    # Handels the key press for Run FastQC action
    def action_run_fastqc(self) -> None:
        self.call_run_FastQC()

    # Handels the key press for Run MultiQC action
    def action_run_multiqc(self) -> None:
        self.call_run_MultiQC()

    # Handels the key press for Run alignment action
    def action_run_alignment(self) -> None:
        self.call_run_bwa_mem()

    # Handels the key press for Run index action
    def action_run_index(self) -> None:
        self.call_run_bwa_index()

    # Handels the key press for Show help action
    def action_show_help(self) -> None:
        self.query_one(TabbedContent).active = "HelpTab"

    # Handels the key press for Exit application action
    def action_exit_app(self) -> None:
        self.push_screen(
            YesOrNo("Do you want to exit application?"), self.maybe_exit_app
        )

    # Confirm exit application
    def maybe_exit_app(self, bool) -> None:
        if bool:
            self.exit()

    # Update the full output path based on current state
    def update_full_output_path(self) -> None:
        self.full_output_path = (
            f"{self.workingDir}/data/{self.outputdir_name}/{self.outputfile_name}"
        )

    # Watcher for full output path changes
    def watch_full_output_path(self, full_output_path: str) -> None:
        self.query_one("#output_path_label", Label).update(
            f"Saved in: {full_output_path}"
        )


# Main entry point of the application
def main():
    logging.info("Application started.")
    Varcall().run()
    logging.info("Application closed.")


if __name__ == "__main__":
    main()



import filecmp
import unittest
import pytest
from varcall.main import Varcall
from textual.pilot import Pilot

app = Varcall()

def test_hometab(snap_compare):
    assert snap_compare("../src/varcall/main.py", press=["tab", "right"])

def test_samtoolstab(snap_compare):
    assert snap_compare("../src/varcall/main.py", press=["tab", "right", "right"])

def test_bcftools(snap_compare):
    assert snap_compare("../src/varcall/main.py", press=["tab", "right", "right", "right"])

def test_pipelinetab(snap_compare):
    assert snap_compare("../src/varcall/main.py", press=["tab", "left"])

def test_helptab(snap_compare):
    assert snap_compare("../src/varcall/main.py")

def test_homebuttons(snap_compare):

    @pytest.mark.asyncio
    async def home_download_button(pilot: Pilot):
        await pilot.resize_terminal(168, 150)
        await pilot.press("tab", "right", "tab", "t", "r", "i", "a", "l", "enter")
        await pilot.click("#Download_button")
    assert snap_compare(Varcall(), run_before=home_download_button,  terminal_size=(168, 75))

    # @pytest.mark.asyncio
    # async def home_fastqc_button(pilot: Pilot):
    #     await pilot.resize_terminal(168, 75)
    #     await pilot.press("tab", "right", "tab", "t", "r", "i", "a", "l", "enter")
    #     await pilot.click("#FastQC_Button")
    # assert snap_compare(Varcall(), run_before=home_fastqc_button,  terminal_size=(168, 75))
    #
    # @pytest.mark.asyncio
    # async def home_multiqc_button(pilot: Pilot):
    #     await pilot.resize_terminal(168, 75)
    #     await pilot.press("tab", "right", "tab", "t", "r", "i", "a", "l", "enter")
    #     await pilot.click("#MultiQC_Button")
    # assert snap_compare(Varcall(), run_before=home_multiqc_button,  terminal_size=(168, 75))

    @pytest.mark.asyncio
    async def home_bwa_index_button(pilot: Pilot):
        await pilot.resize_terminal(168, 75)
        await pilot.press("tab", "right", "tab", "t", "r", "i", "a", "l", "enter")
        await pilot.click("#bwa_index_Button")
    assert snap_compare(Varcall(), run_before=home_bwa_index_button,  terminal_size=(168, 75))

    @pytest.mark.asyncio
    async def home_bwa_mem_button(pilot: Pilot):
        await pilot.resize_terminal(168, 75)
        await pilot.press("tab", "right", "tab", "t", "r", "i", "a", "l", "enter")
        await pilot.click("#bwa_mem_Button")
    assert snap_compare(Varcall(), run_before=home_bwa_mem_button,  terminal_size=(168, 75))

def test_pipeline_buttons(snap_compare):

    @pytest.mark.asyncio
    async def pipeline_button(pilot: Pilot):
        await pilot.resize_terminal(168, 40)
        await pilot.press("tab", "left", "end")
        await pilot.click("#start_pipeline_button")
    assert snap_compare(Varcall(), run_before=pipeline_button,  terminal_size=(168, 40))

class TestFindMatchingFiles(unittest.TestCase):
    def compare_file(self):
        test_files_dir = "./trial/" #created manually by testing the app
        template_files_dir = "./test_trial" # created by main_test.sh
        _, mismatch, _ = filecmp.cmpfiles(test_files_dir, template_files_dir, shallow=False, common=["t"])
        assert mismatch == []

if __name__ == "__main__":
    test_samtoolstab(Varcall)
    test_hometab(Varcall)



from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.widgets import Button, Static, Label, Input

from varcall.modules.exec.common  import file_suggester


class BcftoolsWidgets(Static):
    def compose(self) -> ComposeResult:

        with Container(id="bcf_mpileup_container"):
            yield Label("bcftools mpileup")
            with Horizontal(id="bcf_mpileup_horizontal"):
                yield Input( placeholder="reference", id="bcf_mpileup_reference_input", suggester=file_suggester,)
                yield Input( placeholder="input bcf file name", id="bcf_mpileup_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bcf file name", id="bcf_mpileup_output_input", suggester=file_suggester,)
                yield Button( "mpileup", id="bcf_mpileup_button",)
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]Generate VCF or BCF containing genotype likelihoods for bcf or CRAM files.", classes="description")

        with Container(id="bcf_call_container"):
            yield Label("bcftools call")
            with Horizontal(id="bcf_call_horizontal"):
                yield Input( placeholder="input bcf/vcf file name", id="bcf_call_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bcf/vcf file name", id="bcf_call_output_input", suggester=file_suggester,)
                yield Button("call bcf file", id="bcf_call_button")
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]compiles the genotype likelihoods and calls only the genetic variants", classes="description")

        with Container(id="bcf_filter_container"):
            yield Label("bcftools filter")
            with Horizontal(id="bcf_filter_horizontal"):
                yield Input( placeholder="filter", id="bcf_filter_filter_input", suggester=file_suggester,)
                yield Input( placeholder="input bcf file name", id="bcf_filter_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bcf file name", id="bcf_filter_output_input", suggester=file_suggester,)
                yield Button( "filter", id="bcf_filter_button",)
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]Apply filters to the called bcf/vcf data", classes="description")

        with Container(id="bcf_norm_container"):
            yield Label("bcftools norm")
            with Horizontal(id="bcf_norm_horizontal"):
                yield Input( placeholder="input bcf/bcf file name", id="bcf_norm_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bcf/bcf file name", id="bcf_norm_output_input", suggester=file_suggester,)
                yield Button("norm bcf file", id="bcf_call_button")
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]Normalise the data", classes="description")

        with Container(id="bcf_stats_container"):
            yield Label("bcftools stats")
            with Horizontal(id="bcf_stats_horizontal"):
                yield Input( placeholder="input bcf/bcf file name", id="bcf_stats_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bcf/bcf file name", id="bcf_stats_output_input", suggester=file_suggester,)
                yield Button("stats bcf file", id="bcf_stats_button")
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]Get the stats of bvf/vcf files", classes="description")

# TODO:

"""
# Bcftools Commands:
1. bcftools mpileup    # Generate VCF/BCF files from alignment files.
2. bcftools call       # Call variants (SNPs/indels).
3. bcftools filter     # Apply filters to VCF/BCF files.
4. bcftools norm       # Normalize variants (e.g., left-align indels).
5. bcftools merge      # Merge multiple VCF/BCF files.
6. bcftools annotate   # Annotate VCF files with additional information.
7. bcftools stats      # Generate statistics for VCF/BCF files.
"""



import logging
import subprocess
import threading

from varcall.modules.exec.common  import get_input
from varcall.modules.logging import setup_logging

setup_logging()

# 1. bcftools mpileup command
def run_bcf_mpileup(self):
    """

    """
    if not self.workingDir:
        self.workingDir = __file__
    ref = get_input(self, "bcf_mpileup_reference_input")
    input_path = get_input(self, "bcf_mpileup_input_input")
    output_path = get_input(self, "bcf_mpileup_output_input")
    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools mpileup")
        return
    if not output_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools mpileup")
        return
    if not ref:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools mpileup")
        return

    self.notify(f"bcf mpileup {str(input_path)}...", title="bcf mpileup")
    logging.info(f"bcf mpileup {str(input_path)}...")
    self.query_one("#bcf_mpileup_horizontal").add_class("running")

    threading.Thread(
        target=_run_bcf_mpileup,
        args=(
            self,
            ref,
            input_path,
            output_path,
        ),
    ).start()


def _run_bcf_mpileup(self, ref: str, input_path: str, output_path: str) -> None:
    """

    """
    try:
        bcf_mpileup_cmd  = f"bcftools mpileup -f {ref} {input_path} -o {output_path}"
        self.notify(bcf_mpileup_cmd, title="bcf mpileup")
        logging.info("Running command: " + bcf_mpileup_cmd)
        subprocess.run(
            bcf_mpileup_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bcf mpileup completed for {str(input_path)}")
        self.notify(f"bcf mpileup completed for {str(input_path)}", title="bcf mpileup")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bcf mpileup: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bcf mpileup",
        )

    finally:
        self.query_one("#bcf_mpileup_horizontal").remove_class("running")


# 2. bcftools call command

def run_bcf_call(self):
    """

    """
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "bcf_call_input_input")
    output_path = get_input(self, "bcf_call_output_input")
    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools call")
        return
    if not output_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools call")
        return

    self.notify(f"bcf call {str(input_path)}...", title="bcf call")
    logging.info(f"bcf call {str(input_path)}...")
    self.query_one("#bcf_call_horizontal").add_class("running")
    threading.Thread(
        target=_run_bcf_call,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()


def _run_bcf_call(self, input_path: str, output_path: str) -> None:
    """

    """
    try:
        bcf_call_cmd = f"bcftools call -mv -Oz -o {output_path} {input_path}"
        self.notify(bcf_call_cmd, title="bcf call")
        logging.info("Running command: " + bcf_call_cmd)
        subprocess.run(
            bcf_call_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bcf call completed for {str(input_path)}")
        self.notify(f"bcf call completed for {str(input_path)}", title="bcf call")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bcf call: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bcf call",
        )

    finally:
        self.query_one("#bcf_call_horizontal").remove_class("running")

# 3. bcftools filter command
def run_bcf_filter(self):
    """

    """
    if not self.workingDir:
        self.workingDir = __file__
    filter = get_input(self, "bcf_filter_filter_input")
    input_path = get_input(self, "bcf_filter_input_input")
    output_path = get_input(self, "bcf_filter_output_input")

    if not filter:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools filter")
        return
    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools filter")
        return
    if not output_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools filter")
        return

    self.notify(f"bcf filter {str(input_path)}...", title="bcf filter")
    logging.info(f"bcf filter {str(input_path)}...")
    self.query_one("#bcf_filter_horizontal").add_class("running")
    threading.Thread(
        target=_run_bcf_filter,
        args=(
            self,
            filter,
            input_path,
            output_path,
        ),
    ).start()


def _run_bcf_filter(self,filter: str, input_path: str, output_path: str) -> None:
    """

    """
    try:
        bcf_filter_cmd = f"bcftools filter -s LOWQUAL -e '{filter}' {input_path} -o {output_path}"
        self.notify(bcf_filter_cmd, title="bcf filter")
        logging.info("Running command: " + bcf_filter_cmd)
        subprocess.run(
            bcf_filter_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bcf filter completed for {str(input_path)}")
        self.notify(f"bcf filter completed for {str(input_path)}", title="bcf filter")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bcf filter: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bcf filter",
        )

    finally:
        self.query_one("#bcf_filter_horizontal").remove_class("running")



# 4. bcftools norm command

def run_bcf_norm(self):
    """

    """
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "bcf_norm_input_input")
    output_path = get_input(self, "bcf_norm_output_input")
    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools norm")
        return
    if not output_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools norm")
        return

    self.notify(f"bcf norm {str(input_path)}...", title="bcf norm")
    logging.info(f"bcf norm {str(input_path)}...")
    self.query_one("#bcf_norm_horizontal").add_class("running")
    threading.Thread(
        target=_run_bcf_norm,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()


def _run_bcf_norm(self, input_path: str, output_path: str) -> None:
    """

    """
    try:
        bcf_norm_cmd = f"bcftools norm -f {input_path} {output_path}"
        self.notify(bcf_norm_cmd, title="bcf norm")
        logging.info("Running command: " + bcf_norm_cmd)
        subprocess.run(
            bcf_norm_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bcf norm completed for {str(input_path)}")
        self.notify(f"bcf norm completed for {str(input_path)}", title="bcf norm")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bcf norm: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bcf norm",
        )

    finally:
        self.query_one("#bcf_norm_horizontal").remove_class("running")


# 5. bcftools stats command

def run_bcf_stats(self):
    """

    """
    if not self.workingDir:
        self.workingDir = __file__
    input_path = get_input(self, "bcf_stats_input_input")
    output_path = get_input(self, "bcf_stats_output_input")
    if not input_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools stats")
        return
    if not output_path:
        self.notify("Please provide a valid path", severity="warning", title="Bcftools stats")
        return

    self.notify(f"bcf stats {str(input_path)}...", title="bcf stats")
    logging.info(f"bcf stats {str(input_path)}...")
    self.query_one("#bcf_stats_horizontal").add_class("running")
    threading.Thread(
        target=_run_bcf_stats,
        args=(
            self,
            input_path,
            output_path,
        ),
    ).start()


def _run_bcf_stats(self, input_path: str, output_path: str) -> None:
    """

    """
    try:
        bcf_stats_cmd = f"bcftools stats {input_path} > {output_path}"
        self.notify(bcf_stats_cmd, title="bcf stats")
        logging.info("Running command: " + bcf_stats_cmd)
        subprocess.run(
            bcf_stats_cmd.split(),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        logging.info(f"bcf stats completed for {str(input_path)}")
        self.notify(f"bcf stats completed for {str(input_path)}", title="bcf stats")
    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred during bcf stats: {e.stderr}")
        self.notify(
            f"An error occurred during FastQc: {e.stderr}",
            severity="error",
            timeout=10.0,
            title="bcf stats",
        )

    finally:
        self.query_one("#bcf_stats_horizontal").remove_class("running")



from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.validation import Number
from textual.widgets import Button, Input, Label, LoadingIndicator, Select, Static

from varcall.modules.exec.common  import file_suggester


class HomeWidgets(Static):
    selection_list = [
        "reference",
        "reads",
        "vcf",
        "bed",
        "gtf",
        "gff",
    ]

    def compose(self) -> ComposeResult:
        with Horizontal(id="InputProjectName"):
            yield Label( "Enter project name (Absolute path is preferred)", id="Project_name")
            yield Input(id="Input_Project_name")
            yield Label( "[bold #01B0DC]This will be your working directory & is necessary to run all the processes[/bold #01B0DC]", id="Project_location")

        with Container(id="Download_widget"):
            yield Label("Download (uses curl)", id="download_title")

            with Horizontal(id="Download_inputs"):
                yield Input(placeholder="Input url", id="Input_url")
                yield Input( placeholder="outputfile name", id="Input_outputfile_name", suggester=file_suggester)

            with Horizontal(id="Download_options"):
                yield Select.from_values( self.selection_list, prompt="The downloaded file is for", id="Select_outputdir")
                yield Label("Saved in: ", id="output_path_label")
                yield Button("Download", id="Download_button", classes="action_buttons")
                yield LoadingIndicator(id="Download_loading")

        with Container(id="FastQC_widget"):
            yield Label("FastQC", id="FastQC_title")
            with Horizontal(id="MultiQC_horizontal"):
                yield Input( placeholder="Input file name", id="FastQC_input_input", suggester=file_suggester)
                yield Input( placeholder="Output file name", id="FastQC_output_input", suggester=file_suggester)

            with Horizontal(id="FastQC_Horizontal"):
                yield Label( "FastQCs the files in the workingdir/data/reads directory [u]if the input field is left empty[/u]", id="bwa_description")
                yield Button("Fastqc", id="FastQC_Button", classes="action_buttons")
                yield Button("View Results", id="view_FastQC_res")
                yield LoadingIndicator(id="FastQC_Loading")

        with Container(id="MultiQC_widget"):
            yield Label("MultiQC", id="MultiQC_title")
            with Horizontal(id="MultiQC_horizontal"):
                yield Input( placeholder="Input file name", id="MultiQC_input_input", suggester=file_suggester)
                yield Input( placeholder="Output file name", id="MultiQC_output_input", suggester=file_suggester)

            with Horizontal(id="MultiQC_Horizontal"):
                yield Label(
                    "MultiQCs the files in the workingdir/results/fastqc directory [u]if the input field is left empty[/u] \n[bold red]NOTE[/bold red]: MultiQC can be used to visualise the output of various tools like samtools flagstats.\nFor more info check this",
                    id="bwa_description",
                )
                yield Button("MultiQC", id="MultiQC_Button", classes="action_buttons")
                yield Button("View Results", id="view_MultiQC_res")
                yield LoadingIndicator(id="MultiQC_Loading")

        with Container(id="bwa_widget"):
            yield Label("BWA (alignment uses mem)", id="bwa_title")

            with Horizontal(id="bwa_input_horizontal_1"):
                yield Input( placeholder="Input reference genome file name", id="bwa_ref_input", suggester=file_suggester)
                yield Input( placeholder="No. of threads (default=4)", id="bwa_threads_input", validators=[Number(minimum=1, maximum=500)])

            with Horizontal(id="bwa_input_horizontal_2"):
                yield Input( placeholder="Input read file", id="bwa_reads_input_1", suggester=file_suggester)
                yield Input( placeholder="Input read file", id="bwa_reads_input_2", suggester=file_suggester)

            yield Input( placeholder="Output file name", id="bwa_output_input", suggester=file_suggester)

            with Horizontal(id="bwa_Horizontal"):
                yield Label(
                    "[bold red]NOTE[/bold red]: The reference genome must be inedx before mapping\nIndexes all files in workingdir/data/reference directory if the input field is left empty\nMake sure to input two reads (1 sample)",
                    id="bwa_description",
                )
                yield Button("Index", id="bwa_index_Button", classes="action_buttons")
                yield Button("Align", id="bwa_mem_Button", classes="action_buttons")
                yield Button("View Results", id="view_bwa_res")
                yield LoadingIndicator(id="bwa_Loading")



from textual.app import ComposeResult
from textual.widgets import Button, Input, Label, Static
from textual.containers import Container, Horizontal
from varcall.modules.exec.samtools  import *  # noqa: F403
from varcall.modules.exec.common  import file_suggester
from varcall.modules.logging import setup_logging

setup_logging()


class SamtoolsWidgets(Static):

    def compose(self) -> ComposeResult:

        with Container(id="sam_view_container"):
            yield Label("Samtools view")
            with Horizontal(id="sam_view_horizontal"):
                yield Input( placeholder="input sam/bam file name", id="sam_view_input_input", suggester=file_suggester,)
                yield Input( placeholder="output bam file name", id="sam_view_output_input", suggester=file_suggester,)
                yield Input( placeholder="region (optional)", id="sam_view_region_input", suggester=file_suggester,)
                yield Button( "Convert sam to bam", id="sam_view_button",)
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]Convert .sam files to .bam files. you can also select only to convert a certain region", classes="description")

        with Container(id="sam_sort_container"):
            yield Label("Samtools sort")
            with Horizontal(id="sam_sort_horizontal"):
                yield Input( placeholder="input sam/bam file name", id="sam_sort_input_input", suggester=file_suggester,)
                yield Input( placeholder="output sam/bam file name", id="sam_sort_output_input", suggester=file_suggester,)
                yield Button("sort bam file", id="sam_sort_button")
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC].sam and .bam files need to be sorted before futher processing", classes="description")

        with Container(id="sam_index_container"):
            yield Label("Samtools index")
            with Horizontal(id="sam_index_horizontal"):
                yield Input( placeholder="input bam file name", id="sam_index_input_input", suggester=file_suggester,)
                yield Button("index bam file", id="sam_index_button")
            yield Label("[bold #01B0DC]Description: [/bold #01B0DC]index the .bam file for futher analysis", classes="description")



from textual import on
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widgets import Button, Input, Label, Static


class PipelineWidgets(Static):
    def compose(self) -> ComposeResult:
        with Vertical(id="Pipeline_container"):
            yield Label( "[bold red]NOTE: [/bold red]It is recommended to do fastqc and make sure the reads are of sufficiant quality before starting the pipeline")

            yield Input(placeholder="Enter project name", id="pipeline_project_name_input")
            yield Input(placeholder="Enter reference file", id="pipeline_ref_input")
            yield Input(placeholder="Enter a read file", id="pipeline_read1_input")
            yield Input(placeholder="Enter the other read file", id="pipeline_read2_input")

            yield Button("Start Pipeline", id="start_pipeline_button", variant="success")

    @on(Button.Pressed, "#start_pipeline_button")
    def start_pipeline(self):
        self.notify("Not yet implemented", title="Pipeline", timeout=10.0)



import os.path
from pathlib import Path
from textual.widgets import Markdown
from textual.app import ComposeResult


# Function to read the Markdown file
def read_markdown_file(file_path: str) -> str:
    with open(file_path, "r") as file:
        return file.read()


# Specify the path to the Help.md file
path = os.path.dirname( __file__ )
help_markdown_path = Path(f"{path}/Help.md")

# Read the content of the Markdown file
help_markdown_content = read_markdown_file(help_markdown_path)


class HelpMarkdown(Markdown):
    def compose(self) -> ComposeResult:
        yield Markdown(help_markdown_content)
